import glob
import hashlib
import os
import shutil
import site
import sys
import warnings

import metatensor.torch
import torch

from . import _c_lib


METATOMIC_TORCH_LIB_PATH = _c_lib._lib_path()
METATENSOR_TORCH_LIB_PATH = metatensor.torch._c_lib._lib_path()


def _rascaline_lib_path():
    # This is kept for backward compatibility, but rascaline is now named featomic.
    # This code should be removed by the middle of 2025.
    import rascaline

    return [rascaline._c_lib._lib_path()]


def _featomic_deps_path():
    import featomic

    deps_path = [featomic._c_lib._lib_path()]

    libgomp_path = _find_openmp_dep("featomic_torch.libs")
    if libgomp_path is not None:
        deps_path.insert(0, libgomp_path)

    return deps_path


def _sphericart_deps_path():
    import sphericart.torch

    deps_path = []

    libgomp_path = _find_openmp_dep("sphericart_torch.libs")
    if libgomp_path is not None:
        deps_path.append(libgomp_path)

    if sys.platform.startswith("linux"):
        # sphericart uses a separate library to get the CUDA stream corresponding to a
        # tensor, see https://github.com/lab-cosmo/sphericart/pull/164
        sphericart_torch_path = sphericart.torch._lib_path()
        lib_dir = os.path.dirname(sphericart_torch_path)

        cuda_stream_lib = os.path.join(lib_dir, "libsphericart_torch_cuda_stream.so")
        if os.path.exists(cuda_stream_lib):
            deps_path.append(cuda_stream_lib)

    return deps_path


def _find_openmp_dep(search_dir):
    """
    When building code that uses OpenMP on linux, we typically dynamically link to
    libgomp. `cibuildwheel` then copies ``libgomp.so`` to
    ``<wheel_name>.libs/libgomp-<hash>.so``, so we need to find and add this shared
    library to the extensions dependencies.
    """
    if sys.platform.startswith("linux"):
        libs_list = []

        for prefix in site.getsitepackages():
            libs_dir = os.path.join(prefix, search_dir)
            if os.path.exists(libs_dir):
                libs_list = glob.glob(os.path.join(libs_dir, "libgomp-*.so*"))
                if len(libs_list) != 0:
                    # found it!
                    break

        if len(libs_list) == 0:
            warnings.warn(
                f"No libgomp shared library found in '{search_dir}'. "
                "This may cause issues when loading and running the model.",
                stacklevel=2,
            )
        elif len(libs_list) > 1:
            raise RuntimeError(
                f"Multiple libgomp shared libraries found in '{search_dir}': "
                f"{libs_list}. Try to re-install in a fresh environment."
            )
        else:  # len(libs_list) == 1
            return libs_list[0]


# Manual definition of which TorchScript extensions have their own dependencies. The
# dependencies should be returned in the order they need to be loaded.
EXTENSIONS_WITH_DEPENDENCIES = {
    "rascaline_torch": _rascaline_lib_path,
    "featomic_torch": _featomic_deps_path,
    "sphericart_torch": _sphericart_deps_path,
}


def _collect_extensions(extensions_path):
    """
    Record the list of loaded TorchScript extensions (and their dependencies), to check
    that they are also loaded when executing the model.
    """
    if extensions_path is not None:
        if os.path.exists(extensions_path):
            shutil.rmtree(extensions_path)
        os.makedirs(extensions_path)
        # TODO: the extensions are currently collected in a separate directory,
        # should we store the files directly inside the model file? This would makes
        # the model platform-specific but much more convenient (since the end user
        # does not have to move a model around)

    extensions = []
    extensions_deps = []
    for library in torch.ops.loaded_libraries:
        assert os.path.exists(library)

        # these should be provided by the simulation engine
        if os.path.samefile(library, METATENSOR_TORCH_LIB_PATH):
            continue
        if os.path.samefile(library, METATOMIC_TORCH_LIB_PATH):
            continue

        path = _copy_extension(library, extensions_path)
        info = _extension_info(library)
        info["path"] = path
        extensions.append(info)

        for extra in EXTENSIONS_WITH_DEPENDENCIES.get(info["name"], lambda: [])():
            path = _copy_extension(extra, extensions_path)
            info = _extension_info(extra)
            info["path"] = path
            extensions_deps.append(info)

    return extensions, extensions_deps


def _copy_extension(full_path, extensions_dir):
    full_path = os.path.realpath(full_path)

    prefixes = site.getsitepackages()
    if site.ENABLE_USER_SITE:
        prefixes.append(site.getusersitepackages())

    # this also takes care of extensions installed directly in the same prefix as
    # Python, for example when installing a standalone libmetatensor_torch with conda.
    prefixes.append(sys.prefix)

    path = full_path
    for prefix in prefixes:
        prefix = os.path.realpath(prefix)
        assert os.path.isabs(prefix)

        # Remove any local prefix
        if path.startswith(prefix):
            path = os.path.relpath(path, prefix)
            break

    if extensions_dir is not None:
        collect_path = os.path.realpath(os.path.join(extensions_dir, path))
        if collect_path == path:
            raise RuntimeError(
                f"extensions directory '{extensions_dir}' would overwrite files, "
                "you should set it to a local path instead"
            )

        if os.path.exists(collect_path):
            raise RuntimeError(
                f"more than one extension would be collected at {collect_path}"
            )

        os.makedirs(os.path.dirname(collect_path), exist_ok=True)
        shutil.copyfile(full_path, collect_path)

    return path


def _extension_info(path):
    # get the name of the library, excluding any shared object prefix/suffix
    name = os.path.basename(path)
    if name.startswith("lib"):
        name = name[3:]

    if name.endswith(".so"):
        name = name[:-3]

    if name.endswith(".dll"):
        name = name[:-4]

    if name.endswith(".dylib"):
        name = name[:-6]

    # Collect the hash of the extension shared library. We don't currently use
    # this, but it would allow for binary-level reproducibility later.
    with open(path, "rb") as fd:
        sha256 = hashlib.sha256(fd.read()).hexdigest()

    return {"name": name, "sha256": sha256}



import datetime
import json
import math
import os
import platform
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Union

import torch
from metatensor.torch import Labels, TensorBlock, TensorMap, dtype_name
from torch.profiler import record_function

from . import (
    ModelCapabilities,
    ModelEvaluationOptions,
    ModelMetadata,
    ModelOutput,
    NeighborListOptions,
    System,
    check_atomistic_model,
    load_model_extensions,
    unit_conversion_factor,
)
from . import __version__ as metatomic_version
from ._extensions import _collect_extensions
from .outputs import _check_outputs


def load_atomistic_model(path, extensions_directory=None) -> "AtomisticModel":
    """
    Check and then load the metatomic model at the given `path`.

    This function calls :py:func:`metatomic.torch.check_atomistic_model()` and
    :py:func:`metatomic.torch.load_model_extensions()` before attempting to load the
    model.

    :param path: path to an exported metatensor model
    :param extensions_directory: path to a directory containing all extensions required
        by the exported model
    """
    path = str(path)
    load_model_extensions(path, str(extensions_directory))
    check_atomistic_model(path)

    model = torch.jit.load(path)
    return AtomisticModel(model, model.metadata(), model.capabilities())


def is_atomistic_model(module: torch.nn.Module) -> bool:
    """
    Check if a loaded model is an atomistic model.

    :param model: model to check
    :raises TypeError: if the model is not a :py:class:`torch.nn.Module`
    :return: ``True`` if the ``model`` has been exported, ``False`` otherwise.
    """
    if not isinstance(module, torch.nn.Module):
        raise TypeError(
            f"`module` should be a torch.nn.Module, not {type(module).__name__}"
        )

    if isinstance(module, AtomisticModel):
        return True
    elif (
        isinstance(module, torch.jit._script.RecursiveScriptModule)
        and module.original_name == "AtomisticModel"
    ):
        return True
    else:
        return False


class ModelInterface(torch.nn.Module):
    """
    Interface for models that can be used with :py:class:`AtomisticModel`.

    There are several requirements that models must satisfy to be usable with
    :py:class:`AtomisticModel`. The main one is concerns the :py:meth:`forward`
    function, which must have the signature defined in this interface.

    Additionally, the model can request neighbor lists to be computed by the simulation
    engine, and stored inside the input :py:class:`System`. This is done by defining the
    optional :py:meth:`requested_neighbor_lists` method for the model or any of it's
    sub-module.

    :py:class:`AtomisticModel` will check if ``requested_neighbor_lists`` is defined for
    all the sub-modules of the model, then collect and unify identical requests for the
    simulation engine.
    """

    def __init__():
        """"""
        pass

    def forward(
        self,
        systems: List[System],
        outputs: Dict[str, ModelOutput],
        selected_atoms: Optional[Labels],
    ) -> Dict[str, TensorMap]:
        """
        This method should run the model for the given ``systems``, returning the
        requested ``outputs``. If ``selected_atoms`` is a set of :py:class:`Labels`,
        only the corresponding atoms should be included as "main" atoms in the
        calculation and the output.

        ``outputs`` will be a subset of the capabilities that where declared when
        exporting the model. For example if a model can compute both an ``"energy"`` and
        a ``"charge"`` output, the simulation engine might only request one them.

        The returned dictionary should have the same keys as ``outputs``, and the values
        should contains the corresponding properties of the ``systems``, as computed for
        the subset of atoms defined in ``selected_atoms``. Some outputs are
        standardized, and have additional constrains on how the associated metadata
        should look like, documented in the :ref:`atomistic-models-outputs` section. If
        you want to define a new output for your own usage, it name should looks like
        ``"<domain>::<output>"``, where ``<domain>`` indicates who defines this new
        output and ``<output>`` describes the output itself. For example,
        ``"my-package::foobar"`` for a ``foobar`` output defined in ``my-package``.

        The main use case for ``selected_atoms`` is domain decomposition, where the
        :py:class:`System` given to a model might contain both atoms in the current
        domain and some atoms from other domains; and the calculation should produce
        per-atom output only for the atoms in the domain (but still accounting for atoms
        from the other domains as potential neighbors).

        :param systems: atomistic systems on which to run the calculation
        :param outputs: set of outputs requested by the simulation engine
        :param selected_atoms: subset of atoms that should be included in the output,
            defaults to None
        :return: properties of the systems, as predicted by the machine learning model
        """

    def requested_neighbor_lists(self) -> List[NeighborListOptions]:
        """
        Optional method declaring which neighbors list this model requires.

        This method can be defined on either the root model or any of it's sub-modules.
        A single module can request multiple neighbors list simultaneously if it needs
        them.

        It is then the responsibility of the code calling the model to:

        1. call this method (or more generally
           :py:meth:`AtomisticModel.requested_neighbor_lists`) to get the list
           of the requested neighbor lists.;
        2. compute all neighbor lists corresponding to these requests and add them to
           the systems before calling the model.
        """


class AtomisticModel(torch.nn.Module):
    """
    :py:class:`AtomisticModel` is the main entry point for atomistic machine learning
    based on metatensor. It is the interface between custom, user-defined models and
    simulation engines. Users should export their models with this class, and use
    :py:meth:`save()` to save the exported model to a file. The exported models can then
    be loaded by a simulation engine to compute properties of atomistic systems.

    When exporting a ``module``, you should declare what the model is capable of (using
    :py:class:`ModelCapabilities`). This includes what units the model expects as input
    and what properties the model can compute (using :py:class:`ModelOutput`). The
    simulation engine will then ask the model to compute some subset of these properties
    (through a :py:class:`ModelEvaluationOptions`), on all or a subset of atoms of an
    atomistic system.

    The exported module must follow the interface defined by :py:class:`ModelInterface`,
    should not already be compiled by TorchScript, and should be in "eval" mode (i.e.
    ``module.training`` should be ``False``).

    For example, a custom module predicting the energy as a constant value times the
    number of atoms could look like this

    >>> class ConstantEnergy(torch.nn.Module):
    ...     def __init__(self, constant: float):
    ...         super().__init__()
    ...         self.constant = torch.tensor(constant).reshape(1, 1)
    ...
    ...     def forward(
    ...         self,
    ...         systems: List[System],
    ...         outputs: Dict[str, ModelOutput],
    ...         selected_atoms: Optional[Labels] = None,
    ...     ) -> Dict[str, TensorMap]:
    ...         results: Dict[str, TensorMap] = {}
    ...         if "energy" in outputs:
    ...             if outputs["energy"].per_atom:
    ...                 raise NotImplementedError("per atom energy is not implemented")
    ...
    ...             dtype = systems[0].positions.dtype
    ...             energies = torch.zeros(len(systems), 1, dtype=dtype)
    ...             for i, system in enumerate(systems):
    ...                 if selected_atoms is None:
    ...                     n_atoms = len(system)
    ...                 else:
    ...                     n_atoms = len(selected_atoms)
    ...
    ...                 energies[i] = self.constant * n_atoms
    ...
    ...             systems_idx = torch.tensor([[i] for i in range(len(systems))])
    ...             energy_block = TensorBlock(
    ...                 values=energies,
    ...                 samples=Labels(["system"], systems_idx.to(torch.int32)),
    ...                 components=torch.jit.annotate(List[Labels], []),
    ...                 properties=Labels(["energy"], torch.tensor([[0]])),
    ...             )
    ...
    ...             results["energy"] = TensorMap(
    ...                 keys=Labels(["_"], torch.tensor([[0]])),
    ...                 blocks=[energy_block],
    ...             )
    ...
    ...         return results

    Wrapping and exporting this model would then look like this:

    >>> import os
    >>> import tempfile
    >>> from metatomic.torch import AtomisticModel
    >>> from metatomic.torch import (
    ...     ModelCapabilities,
    ...     ModelOutput,
    ...     ModelMetadata,
    ... )
    >>> model = ConstantEnergy(constant=3.141592)
    >>> # put the model in inference mode
    >>> model = model.eval()
    >>> # Define the model capabilities
    >>> capabilities = ModelCapabilities(
    ...     outputs={
    ...         "energy": ModelOutput(
    ...             quantity="energy",
    ...             unit="eV",
    ...             per_atom=False,
    ...             explicit_gradients=[],
    ...         ),
    ...     },
    ...     atomic_types=[1, 2, 6, 8, 12],
    ...     interaction_range=0.0,
    ...     length_unit="angstrom",
    ...     supported_devices=["cpu"],
    ...     dtype="float64",
    ... )
    >>> # define metadata about this model
    >>> metadata = ModelMetadata(
    ...     name="model-name",
    ...     authors=["Some Author", "Another One"],
    ...     # references and long description can also be added
    ... )
    >>> # export the model
    >>> model = AtomisticModel(model, metadata, capabilities)
    >>> # save the exported model to disk
    >>> with tempfile.TemporaryDirectory() as directory:
    ...     model.save(os.path.join(directory, "constant-energy-model.pt"))

    .. py:attribute:: module
        :type: ModelInterface

    The torch module exported by this :py:class:`AtomisticModel`.

    Reading from this attribute is safe, but modifying it is not recommended, unless you
    are familiar with the implementation of the model.
    """

    # Some annotation to make the TorchScript compiler happy
    _requested_neighbor_lists: List[NeighborListOptions]

    def __init__(
        self,
        module: ModelInterface,
        metadata: ModelMetadata,
        capabilities: ModelCapabilities,
    ):
        """
        :param module: The torch module to export.
        :param capabilities: Description of the model capabilities.
        """
        super().__init__()

        if is_atomistic_model(module):
            # module was already checked; take the sub-module as is
            self.module = module.module
        else:
            _check_annotation(module)
            self.module = module

        if module.training:
            raise ValueError("module should not be in training mode")

        # ============================================================================ #

        # recursively explore `module` to get all the requested_neighbor_lists
        self._requested_neighbor_lists = []
        _get_requested_neighbor_lists(
            module,
            self.module.__class__.__name__,
            self._requested_neighbor_lists,
            capabilities.length_unit,
        )
        # ============================================================================ #

        self._metadata = metadata
        self._capabilities = capabilities

        # check that some required capabilities are set
        if capabilities.interaction_range < 0:
            raise ValueError(
                "`capabilities.interaction_range` was not set, "
                "but it is required to run simulations"
            )

        if math.isnan(capabilities.interaction_range):
            raise ValueError(
                "`capabilities.interaction_range` should be a "
                "float between 0 and infinity"
            )

        if len(capabilities.supported_devices) == 0:
            raise ValueError(
                "`capabilities.supported_devices` was not set, "
                "but it is required to run simulations."
            )

        if capabilities.dtype == "":
            raise ValueError(
                "`capabilities.dtype` was not set, "
                "but it is required to run simulations."
            )

        if capabilities.dtype == "float32":
            self._model_dtype = torch.float32
        elif capabilities.dtype == "float64":
            self._model_dtype = torch.float64
        else:
            raise ValueError(f"unknown dtype in capabilities: {capabilities.dtype}")

    @torch.jit.export
    def capabilities(self) -> ModelCapabilities:
        """Get the capabilities of the exported model"""
        return self._capabilities

    @torch.jit.export
    def metadata(self) -> ModelMetadata:
        """Get the metadata of the exported model"""
        return self._metadata

    @torch.jit.export
    def requested_neighbor_lists(self) -> List[NeighborListOptions]:
        """
        Get the neighbors lists required by the exported model or any of the child
        module.
        """
        return self._requested_neighbor_lists

    def forward(
        self,
        systems: List[System],
        options: ModelEvaluationOptions,
        check_consistency: bool,
    ) -> Dict[str, TensorMap]:
        """Run the exported model and return the corresponding outputs.

        Before running the model, this will convert the ``systems`` data from the engine
        unit to the model unit, including all neighbors lists distances.

        After running the model, this will convert all the outputs from the model units
        to the engine units.

        :param systems: input systems on which we should run the model. The systems
            should already contain all neighbors lists corresponding to the options in
            :py:meth:`requested_neighbor_lists()`.
        :param options: options for this run of the model
        :param check_consistency: Should we run additional check that everything is
            consistent? This should be set to ``True`` when verifying a model, and to
            ``False`` once you are sure everything is running fine.

        :return: A dictionary containing all the model outputs
        """

        if check_consistency:
            with record_function("AtomisticModel::check_inputs"):
                _check_inputs(
                    capabilities=self._capabilities,
                    requested_neighbor_lists=self._requested_neighbor_lists,
                    systems=systems,
                    options=options,
                    expected_dtype=self._model_dtype,
                )

        with record_function("AtomisticModel::check_atomic_types"):
            # always (i.e. even if check_consistency=False) check that the atomic types
            # of the system match the one the model supports
            if len(systems) > 0:
                all_types = torch.cat([system.types for system in systems])
                supported_types = torch.tensor(
                    self._capabilities.atomic_types,
                    device=all_types.device,
                    dtype=all_types.dtype,
                )
                unsupported_mask = torch.logical_not(
                    torch.isin(all_types, supported_types)
                )
                if torch.any(unsupported_mask):
                    atom_type = all_types[unsupported_mask][0]
                    raise ValueError(
                        "this model does not support the atomic type "
                        f"'{atom_type.item()}' which is present in the input systems"
                    )

        # convert systems from engine to model units
        with record_function("AtomisticModel::convert_units_input"):
            if self._capabilities.length_unit != options.length_unit:
                conversion = unit_conversion_factor(
                    quantity="length",
                    from_unit=options.length_unit,
                    to_unit=self._capabilities.length_unit,
                )

                systems = _convert_systems_units(
                    systems,
                    conversion,
                    model_length_unit=self._capabilities.length_unit,
                    system_length_unit=options.length_unit,
                )

        # run the actual calculations
        with record_function("Model::forward"):
            outputs = self.module(
                systems=systems,
                outputs=options.outputs,
                selected_atoms=options.selected_atoms,
            )

        if check_consistency:
            with record_function("AtomisticModel::check_outputs"):
                _check_outputs(
                    systems=systems,
                    requested=options.outputs,
                    selected_atoms=options.selected_atoms,
                    outputs=outputs,
                    expected_dtype=self._model_dtype,
                )

        # convert outputs from model to engine units
        with record_function("AtomisticModel::convert_units_output"):
            for name, output in outputs.items():
                declared = self._capabilities.outputs[name]
                requested = options.outputs[name]
                if declared.quantity == "" or requested.quantity == "":
                    continue

                if declared.quantity != requested.quantity:
                    raise ValueError(
                        f"model produces values as '{declared.quantity}' for the "
                        f"'{name}' output, but the engine requested "
                        f"'{requested.quantity}'"
                    )

                conversion = unit_conversion_factor(
                    quantity=declared.quantity,
                    from_unit=declared.unit,
                    to_unit=requested.unit,
                )

                if conversion != 1.0:
                    for block in output.blocks():
                        block.values[:] *= conversion
                        for _, gradient in block.gradients():
                            gradient.values[:] *= conversion

        return outputs

    def export(self, file: str, collect_extensions: Optional[str] = None):
        """Export this model to a file that can then be loaded by simulation engine.

        .. warning::
            :py:meth:`export` is deprecated. Use :py:meth:`save` instead.

        :param file: where to save the model. This can be a path or a file-like object.
        :param collect_extensions: if not None, all currently loaded PyTorch extension
            will be collected in this directory. If this directory already exists, it
            is removed and re-created.
        """
        warnings.warn(
            message="`export()` is deprecated, use `save()` instead",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.save(file, collect_extensions)

    def save(self, file: Union[str, Path], collect_extensions: Optional[str] = None):
        """Save this model to a file that can then be loaded by simulation engine.

        The model will be saved with `requires_grad=False` for all parameters.

        :param file: where to save the model. This can be a path or a file-like object.
        :param collect_extensions: if not None, all currently loaded PyTorch extension
            will be collected in this directory. If this directory already exists, it
            is removed and re-created.
        """
        for parameter in self.parameters():
            parameter.requires_grad = False
        module = self.eval()
        if os.environ.get("PYTORCH_JIT") == "0":
            raise RuntimeError(
                "found PYTORCH_JIT=0 in the environment, "
                "we can not save models without TorchScript"
            )

        try:
            module = torch.jit.script(module)
        except RuntimeError as e:
            raise RuntimeError("could not convert the module to TorchScript") from e

        if self._capabilities.length_unit == "":
            warnings.warn(
                "No length unit was provided for the model.",
                stacklevel=1,
            )

        for name, output in self._capabilities.outputs.items():
            # TODO: coordinate a list of standard outputs needing
            # unit checks, should also be consistent with `outputs.py`
            if name in ["energy", "energy_ensemble", "energy_uncertainty"]:
                if output.unit == "":
                    warnings.warn(
                        f"No units were provided for output {name}.",
                        stacklevel=1,
                    )

        # TODO: can we freeze these?
        # module = torch.jit.freeze(module)

        # Metadata about where and when the model was exported
        export_metadata = {
            "date": datetime.datetime.now(datetime.timezone.utc).isoformat(),
            "platform": platform.machine() + "-" + platform.system(),
            # TODO: user/hostname?
        }

        if collect_extensions is not None:
            export_metadata["extensions_directory"] = str(collect_extensions)

        extensions, deps = _collect_extensions(extensions_path=collect_extensions)

        torch.jit.save(
            module.to("cpu"),  # this allows to torch.jit.load without devices
            str(file),
            _extra_files={
                "torch-version": torch.__version__,
                "metatomic-version": metatomic_version,
                "extensions": json.dumps(extensions),
                "extensions-deps": json.dumps(deps),
                "export-metadata": json.dumps(export_metadata),
                "model-metadata": self._metadata.__getstate__()[0],
            },
        )


def _get_requested_neighbor_lists(
    module: torch.nn.Module,
    module_name: str,
    requested: List[NeighborListOptions],
    length_unit: str,
):
    if hasattr(module, "requested_neighbor_lists"):
        for new_options in module.requested_neighbor_lists():
            new_options.add_requestor(module_name)

            already_requested = False
            for existing in requested:
                if existing == new_options:
                    already_requested = True
                    for requestor in new_options.requestors():
                        existing.add_requestor(requestor)

            if not already_requested:
                if new_options.length_unit not in ["", length_unit]:
                    raise ValueError(
                        f"NeighborsListOptions from {module_name} already have a "
                        f"length unit ('{new_options.length_unit}') which does not "
                        f"match the model length units ('{length_unit}')"
                    )

                new_options.length_unit = length_unit
                requested.append(new_options)

    for child_name, child in module.named_children():
        _get_requested_neighbor_lists(
            module=child,
            module_name=module_name + "." + child_name,
            requested=requested,
            length_unit=length_unit,
        )


def _check_annotation(module: torch.nn.Module):
    # check annotations on forward
    annotations = module.forward.__annotations__
    expected_arguments = [
        "systems",
        "outputs",
        "selected_atoms",
        "return",
    ]

    expected_signature = (
        "`forward(self, "
        "systems: List[System], "
        "outputs: Dict[str, ModelOutput], "
        "selected_atoms: Optional[Labels]"
        ") -> Dict[str, TensorMap]`"
    )

    if list(annotations.keys()) != expected_arguments:
        raise TypeError(
            "`module.forward()` takes unexpected arguments, expected signature is "
            + expected_signature
        )

    if annotations["systems"] != List[System]:
        raise TypeError(
            "`systems` argument must be a list of metatensor atomistic `System`, "
            f"not {annotations['system']}"
        )

    if annotations["outputs"] != Dict[str, ModelOutput]:
        raise TypeError(
            "`outputs` argument must be `Dict[str, ModelOutput]`, "
            f"not {annotations['outputs']}"
        )

    if annotations["selected_atoms"] != Optional[Labels]:
        raise TypeError(
            "`selected_atoms` argument must be `Optional[Labels]`, "
            f"not {annotations['selected_atoms']}"
        )

    if annotations["return"] != Dict[str, TensorMap]:
        raise TypeError(
            "`forward()` must return a `Dict[str, TensorMap]`, "
            f"not {annotations['return']}"
        )


def _check_inputs(
    capabilities: ModelCapabilities,
    requested_neighbor_lists: List[NeighborListOptions],
    systems: List[System],
    options: ModelEvaluationOptions,
    expected_dtype: torch.dtype,
):
    if len(systems) == 0:
        return

    global_device = systems[0].device
    global_dtype = systems[0].positions.dtype

    if global_dtype != expected_dtype:
        raise ValueError(
            f"wrong dtype for the data: the model wants {dtype_name(expected_dtype)}, "
            f"we got {dtype_name(global_dtype)}"
        )

    # check that the requested outputs match what the model can do
    for name, requested in options.outputs.items():
        if name not in capabilities.outputs:
            raise ValueError(
                f"this model can not compute '{name}', the implemented "
                f"outputs are {capabilities.outputs.keys()}"
            )

        possible = capabilities.outputs[name]

        for parameter in requested.explicit_gradients:
            if parameter not in possible.explicit_gradients:
                raise ValueError(
                    f"this model can not compute explicit gradients of '{name}' "
                    f"with respect to '{parameter}'"
                )

        if requested.per_atom and not possible.per_atom:
            raise ValueError(
                f"this model can not compute '{name}' per atom, only globally"
            )

    selected_atoms = options.selected_atoms
    if selected_atoms is not None:
        if selected_atoms.device != global_device:
            raise ValueError(
                "expected all selected_atoms to be on the same device as the systems, "
                f"got {selected_atoms.device} and {global_device}"
            )

        if selected_atoms.names != ["system", "atom"]:
            raise ValueError(
                "invalid names for selected_atoms: expected "
                f"['system', 'atom'], got {selected_atoms.names}"
            )

        possible_atoms_values: List[List[int]] = []
        for s, system in enumerate(systems):
            for a in range(len(system)):
                possible_atoms_values.append([s, a])

        possible_atoms = Labels(
            ["system", "atom"],
            torch.tensor(possible_atoms_values, device=global_device),
        )

        intersection = selected_atoms.intersection(possible_atoms)
        if len(intersection) != len(selected_atoms):
            raise ValueError(
                "invalid selected_atoms: there are entries that are not "
                "possible for the current systems"
            )

    for system in systems:
        if system.device != global_device:
            raise ValueError(
                "expected all systems to be on the same device, "
                f"got {global_device} and {system.device}"
            )

        if not system.positions.dtype == global_dtype:
            raise ValueError(
                "expected all systems to have the same dtype, "
                f"got {global_dtype} and {system.positions.dtype}"
            )

        # Check neighbors lists
        known_neighbor_lists = system.known_neighbor_lists()
        for request in requested_neighbor_lists:
            found = False
            for known in known_neighbor_lists:
                if request == known:
                    found = True

            if not found:
                raise ValueError(
                    "missing neighbors list in the system: the model requested "
                    f"a list for {request}, but it was not computed and stored "
                    "in the system"
                )


def _convert_systems_units(
    systems: List[System],
    conversion: float,
    model_length_unit: str,
    system_length_unit: str,
) -> List[System]:
    if conversion == 1.0:
        return systems

    new_systems: List[System] = []
    for system in systems:
        new_system = System(
            types=system.types,
            positions=conversion * system.positions,
            cell=conversion * system.cell,
            pbc=system.pbc,
        )

        # also update the neighbors list distances
        for request in system.known_neighbor_lists():
            neighbors = system.get_neighbor_list(request)
            new_system.add_neighbor_list(
                request,
                TensorBlock(
                    values=conversion * neighbors.values,
                    samples=neighbors.samples,
                    components=neighbors.components,
                    properties=neighbors.properties,
                ),
            )

        known_data = system.known_data()
        if len(known_data) != 0:
            warnings.warn(
                "the model requires a different length unit "
                f"({model_length_unit}) than the system ({system_length_unit}), "
                f"but we don't know how to convert custom data ({known_data}) "
                "accordingly",
                stacklevel=2,
            )

        for data in known_data:
            new_system.add_data(data, system.get_data(data))

        new_systems.append(new_system)

    return new_systems



from typing import Dict, List, Optional

import torch
from metatensor.torch import Labels, TensorBlock, TensorMap


class System:
    """
    A System contains all the information about an atomistic system; and should be used
    as the input of atomistic models.
    """

    def __init__(
        self,
        types: torch.Tensor,
        positions: torch.Tensor,
        cell: torch.Tensor,
        pbc: torch.Tensor,
    ):
        """
        You can create a :py:class:`System` with ``types``, ``positions`` and ``cell``
        tensors, or convert data from other libraries.

        .. admonition:: Converting data to metatensor `System`

            We provide a way to convert :py:class:`ase.Atoms` instances to
            :py:class:`System` using the :py:func:`systems_to_torch()` function.

            In addition, some external packages provide ways to create
            :py:class:`System` using data from other libraries:

            - `featomic`_ has the :py:func:`featomic.torch.systems_to_torch()`
              function that can convert from ASE, chemfiles and PySCF.

            .. _featomic: https://metatensor.github.io/featomic/latest/index.html

        :param types: 1D tensor of integer representing the particles identity. For
            atoms, this is typically their atomic numbers.

        :param positions: 2D tensor of shape (len(types), 3) containing the Cartesian
            positions of all particles in the system.

        :param cell: 2D tensor of shape (3, 3), describing the bounding box/unit cell of
            the system. Each row should be one of the bounding box vector; and columns
            should contain the x, y, and z components of these vectors (i.e. the cell
            should be given in row-major order). Systems that are not periodic along
            one or more directions should set the corresponding cell vectors to 0.

        :param pbc: tensor containing 3 boolean values, indicating which dimensions are
            periodic along each axis, in the same order as the cell vectors.
        """

    def __len__(self) -> int:
        pass

    @property
    def types(self) -> torch.Tensor:
        """Tensor of 32-bit integers representing the particles identity"""

    @property
    def positions(self) -> torch.Tensor:
        """
        Tensor of floating point values containing the particles cartesian coordinates
        """

    @property
    def cell(self) -> torch.Tensor:
        """Tensor of floating point values containing bounding box/cell of the system"""

    @property
    def pbc(self) -> torch.Tensor:
        """Tensor of boolean values indicating which dimensions are periodic"""

    @property
    def device(self) -> torch.device:
        """get the device of all the arrays stored inside this :py:class:`System`"""

    @property
    def dtype(self) -> torch.dtype:
        """
        get the dtype of all the arrays stored inside this :py:class:`System`

        .. warning::

            Due to limitations in TorchScript C++ extensions, the dtype is returned as
            an integer, which can not be compared with :py:class:`torch.dtype`
            instances. See :py:attr:`TensorBlock.dtype
            <metatensor.torch.TensorBlock.dtype>` for more information.
        """

    def to(
        self,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
    ) -> "System":
        """
        Move all the arrays in this system to the given ``dtype`` and ``device``.

        :param dtype: new dtype to use for all arrays. The dtype stays the same if this
            is set to ``None``.
        :param device: new device to use for all arrays. The device stays the same if
            this is set to ``None``.
        """

    def add_neighbor_list(
        self,
        options: "NeighborListOptions",
        neighbors: TensorBlock,
    ):
        """
        Add a new neighbors list in this system corresponding to the given ``options``.

        The neighbors list should have the following samples: ``"first_atom"``,
        ``"second_atom"``, ``"cell_shift_a"``, ``"cell_shift_b"``, ``"cell_shift_c"``,
        containing the index of the first and second atoms (matching the "atom" sample
        in the positions); and the number of cell vector a/b/c to add to the positions
        difference to get the pair vector.

        The neighbors should also have a single component ``"xyz"`` with values ``[0, 1,
        2]``; and a single property ``"distance"`` with value 0.

        The neighbors values must contain the distance vector from the first to the
        second atom, i.e. ``positions[second_atom] - positions[first_atom] +
        cell_shift_a * cell_a + cell_shift_b * cell_b + cell_shift_c * cell_c``.

        :param options: options of the neighbors list
        :param neighbors: list of neighbors stored in a :py:class:`TensorBlock`
        """

    def get_neighbor_list(
        self,
        options: "NeighborListOptions",
    ) -> TensorBlock:
        """
        Retrieve a previously stored neighbors list with the given ``options``, or throw
        an error if no such neighbors list exists.

        :param options: options of the neighbors list to retrieve
        """

    def known_neighbor_lists(self) -> List["NeighborListOptions"]:
        """
        Get all the neighbors lists options registered with this :py:class:`System`
        """

    def add_data(self, name: str, tensor: TensorMap, override: bool = False):
        """
        Add custom data to this system, stored as :py:class:`TensorBlock`.

        This is intended for experimentation with models that need more data as input,
        and moved into a field of ``System`` later.

        :param name: name of the custom data
        :param tensor: the data to store
        :param override: if ``True``, allow this function to override existing data with
            the same name
        """

    def get_data(self, name: str) -> TensorMap:
        """
        Retrieve custom data stored in this System with the given ``name``, or throw
        an error if no data can be found.

        :param name: name of the custom data to retrieve
        """

    def known_data(self) -> List[str]:
        """
        Get the name of all the custom data registered with this :py:class:`System`
        """


class NeighborListOptions:
    """Options for the calculation of a neighbors list"""

    def __init__(
        self, cutoff: float, full_list: bool, strict: bool, requestor: str = ""
    ):
        """
        :param cutoff: spherical cutoff radius for the neighbors list, in the
            model units
        :param full_list: should the list be a full or half neighbors list
        :param strict: whether the list guarantee to have no pairs farther than cutoff
        :param requestor: who requested this neighbors list, you can add additional
            requestors later using :py:meth:`add_requestor`
        """

    @property
    def cutoff(self) -> float:
        """Spherical cutoff radius for this neighbors list in model units"""

    @property
    def length_unit(self) -> str:
        """
        The unit of length used for the cutoff.

        This is typically set by :py:class:`AtomisticModel` when collecting
        all neighbors list requests.

        The list of possible units is available :ref:`here <known-quantities-units>`.
        """

    def engine_cutoff(self, engine_length_unit: str) -> float:
        """
        Spherical cutoff radius for this neighbors list in engine units.

        The engine must provide the unit it uses for lengths, and the cutoff will
        automatically be converted.
        """

    @property
    def full_list(self) -> bool:
        """
        Should the list be a full neighbors list (contains both the pair ``i->j`` and
        ``j->i``) or a half neighbors list (contains only the pair ``i->j``)
        """

    @property
    def strict(self) -> bool:
        """
        Does the list guarantee to have no pairs beyond the cutoff (strict) or
        can it also have pairs that are farther apart (non strict)
        """

    def requestors(self) -> List[str]:
        """Get the list of modules requesting this neighbors list"""

    def add_requestor(self, requestor: str):
        """
        Add another ``requestor`` to the list of modules requesting this neighbors list
        """

    def __repr__(self) -> str:
        pass

    def __str__(self) -> str:
        pass

    def __eq__(self, other: "NeighborListOptions") -> bool:
        pass

    def __ne__(self, other: "NeighborListOptions") -> bool:
        pass


class ModelOutput:
    """Description of one of the quantity a model can compute."""

    def __init__(
        self,
        quantity: str = "",
        unit: str = "",
        per_atom: bool = False,
        explicit_gradients: List[str] = [],  # noqa B006
    ):
        pass

    @property
    def quantity(self) -> str:
        """
        Quantity of the output (e.g. energy, dipole, …).  If this is an empty string, no
        unit conversion will be performed.

        The list of possible quantities is available :ref:`here
        <known-quantities-units>`.
        """

    @property
    def unit(self) -> str:
        """
        Unit of the output. If this is an empty string, no unit conversion will be
        performed.

        The list of possible units is available :ref:`here <known-quantities-units>`.
        """

    per_atom: bool
    """Is the output defined per-atom or for the overall structure"""

    explicit_gradients: List[str]
    """
    Which gradients should be computed eagerly and stored inside the output
    :py:class:`TensorMap`.
    """


class ModelCapabilities:
    """Description of a model capabilities, i.e. everything a model can do."""

    def __init__(
        self,
        outputs: Dict[str, ModelOutput] = {},  # noqa B006
        atomic_types: List[int] = [],  # noqa B006
        interaction_range: float = -1,
        length_unit: str = "",
        supported_devices: List[str] = [],  # noqa B006
        dtype: str = "",
    ):
        pass

    @property
    def outputs(self) -> Dict[str, ModelOutput]:
        """
        All possible outputs from this model and corresponding settings.

        During a specific run, a model might be asked to only compute a subset of these
        outputs. Some outputs are standardized, and have additional constrains on how
        the associated metadata should look like, documented in the
        :ref:`atomistic-models-outputs` section.

        If you want to define a new output for your own usage, it name should looks like
        ``"<domain>::<output>"``, where ``<domain>`` indicates who defines this new
        output and ``<output>`` describes the output itself. For example,
        ``"my-package::foobar"`` for a ``foobar`` output defined in ``my-package``.
        """

    atomic_types: List[int]
    """which atomic types the model can handle"""

    interaction_range: float
    """
    How far a given atom needs to know about other atoms, in the length unit of the
    model.

    For a short range model, this is the same as the largest neighbors list cutoff. For
    a message passing model, this is the cutoff of one environment times the number of
    message passing steps. For an explicit long range model, this should be set to
    infinity (``float("inf")``/``math.inf``/``torch.inf`` in Python).
    """

    @property
    def length_unit() -> str:
        """
        Unit used by the model for its inputs.

        This applies to the ``interaction_range``, any cutoff in neighbors lists, the
        atoms positions and the system cell.

        The list of possible units is available :ref:`here <known-quantities-units>`.
        """

    @property
    def dtype() -> str:
        """
        The dtype of this model

        This can be ``"float32"`` or ``"float64"``, and must be used by the engine as
        the dtype of all inputs and outputs for this model.
        """

    def engine_interaction_range(self, engine_length_unit: str) -> float:
        """
        Same as :py:attr:`interaction_range`, but in the unit of length used by the
        engine.
        """

    supported_devices: List[str]
    """
    What devices can this model run on? This should only contain the ``device_type``
    part of the device, and not the device number (i.e. this should be ``"cuda"``, not
    ``"cuda:0"``).

    Devices should be ordered in order of preference: the first entry in this list
    should be the best device for this model, and so on.
    """


class ModelEvaluationOptions:
    """
    Options requested by the simulation engine/evaluation code when doing a single model
    evaluation.
    """

    def __init__(
        self,
        length_unit: str = "",
        outputs: Dict[str, ModelOutput] = {},  # noqa B006
        selected_atoms: Optional[Labels] = None,
    ):
        pass

    @property
    def length_unit(self) -> str:
        """
        Unit of lengths the engine uses for the model input.

        The list of possible units is available :ref:`here <known-quantities-units>`.
        """

    outputs: Dict[str, ModelOutput]
    """requested outputs for this run and corresponding settings"""

    @property
    def selected_atoms() -> Optional[Labels]:
        """
        Only run the calculation for a selected subset of atoms.

        If this is set to ``None``, run the calculation on all atoms. If this is a set
        of :py:class:`metatensor.torch.Labels`, it will have two dimensions named
        ``"system"`` and ``"atom"``, containing the 0-based indices of all the atoms in
        the selected subset.
        """


class ModelMetadata:
    """
    Metadata about a specific exported model

    This class implements the ``__str__`` and ``__repr__`` methods, so its
    representation can be easily printed, logged, inserted into other strings, etc.
    """

    def __init__(
        self,
        name: str = "",
        description: str = "",
        authors: List[str] = [],  # noqa: B006
        references: Dict[str, List[str]] = {},  # noqa: B006
        extra: Dict[str, str] = {},  # noqa: B006
    ):
        pass

    name: str
    """Name of this model"""

    description: str
    """Description of this model"""

    authors: List[str]
    """List of authors for this model"""

    references: Dict[str, List[str]]
    """
    Academic references for this model. The top level dict can have three keys:

    - "implementation": for reference to software used in the implementation
      of the model
    - "architecture": for reference that introduced the general architecture
      used by this model
    - "model": for reference specific to this exact model
    """

    extra: Dict[str, str]
    """
    Any additional metadata that is not contained in the other fields. There are
    no constraints on the keys or values of this dictionary. The extra metadata
    is intended to be used by models to store data they need.
    """


def read_model_metadata(path: str) -> ModelMetadata:
    """
    Read metadata of a saved atomistic model.

    This function allows to access the metadata of a model without loading it
    in advance.

    :param path: path to the exported model file
    """


def check_atomistic_model(path: str):
    """
    Check that the file at ``path`` contains an exported atomistic model, and that this
    model can be loaded in the current process.

    This function should be called before :py:func:`torch.jit.load()` when loading an
    existing model.

    :param path: path to the exported model file
    """


def load_model_extensions(path: str, extensions_directory: Optional[str] = None):
    """
    Load the TorchScript extensions (and their dependencies) that the model at ``path``
    uses.

    If ``extensions_directory`` is provided, we look for the extensions and their
    dependencies in there first. If this function fails to load some library, it will
    produce a warning using Torch's warnings infrastructure. Users can set the
    ``METATENSOR_DEBUG_EXTENSIONS_LOADING`` environment variable to get more
    informations about a failure in the standard error output.

    :param path: path to the exported model file
    :param extensions_directory: path to a directory containing the extensions. This
        directory will typically be created by calling
        :py:meth:`AtomisticModel.export` with
        ``collect_extensions=extensions_directory``.
    """


def register_autograd_neighbors(
    system: System, neighbors: TensorBlock, check_consistency: bool
):
    """
    Register a new torch autograd node going from (``system.positions``,
    ``system.cell``) to the ``neighbors`` distance vectors.

    This does not recompute the distance vectors, but work as-if all the data in
    ``neighbors.values`` was computed directly from ``system.positions`` and
    ``system.cell``, allowing downstream models to use it directly with full autograd
    integration.

    :param system: system containing the positions and cell used to compute the
        neighbors list
    :param neighbors: neighbors list, following the same format as
        :py:meth:`System.add_neighbor_list`
    :param check_consistency: can be set to ``True`` to run additional checks in case
        the data in neighbors does not follow what's expected.
    """


def unit_conversion_factor(quantity: str, from_unit: str, to_unit: str):
    """
    Get the multiplicative conversion factor from ``from_unit`` to ``to_unit``. Both
    units must be valid and known for the given physical ``quantity``. The set of valid
    quantities and units is available :ref:`here <known-quantities-units>`.

    :param quantity: name of the physical quantity
    :param from_unit: current unit of the data
    :param to_unit: target unit of the data
    """



import io
import os
import warnings
import zipfile
from pathlib import Path
from typing import Union

import metatensor.torch
import numpy as np
import torch

from . import NeighborListOptions, System


def save(file: Union[str, Path, io.BytesIO], system: System) -> None:
    """Save a System object to a file.

    The provided System must contain float64 data and be on the CPU device.

    The saved file will be a zip archive containing the following files:

    - ``types.npy``, containing the atomic types in numpy's NPY format;
    - ``positions.npy``, containing the systems' positions in numpy's NPY format;
    - ``cell.npy``, containing the systems' cell in numpy's NPY format;
    - ``pbc.npy``, containing the periodic boundary conditions in numpy's NPY format;

    For each neighbor list in the System object, the following files will be saved
    (where ``{nl_idx}`` is the index of the neighbor list):

    - ``pairs/{nl_idx}/options.json``: the ``NeighborListOptions`` object
      converted to a JSON string.
    - ``pairs/{nl_idx}/data.mts``: the neighbor list ``TensorBlock`` object

    For each extra data in the System object, the following file will be saved (where
    ``{name}`` is the name of the extra data):

    - ``data/{name}.mts``: The extra data ``TensorMap``

    :param file: The path (or file-like object) to save the System to.
    :param system: The System object to save.
    """

    if isinstance(file, (str, Path)):
        if not file.endswith(".mta"):
            raise ValueError("The provided path must have the `.mta` extension.")

    if _is_system(system):
        _save_system(file, system)
    else:
        raise ValueError("`system` must be a System object.")


def load_system(file: Union[str, Path, io.BytesIO]) -> System:
    """Load a System object from a file.

    The loaded System object will be on the CPU device and contain float64 data.

    :param file: The path (or file-like object) to load the System object from.
    :return: The System object.
    """
    if isinstance(file, (str, Path)):
        if not os.path.exists(file):
            raise FileNotFoundError(f"File not found: {file}")
        if not file.endswith(".mta"):
            raise ValueError("The provided path must have the `.mta` extension.")

    if _is_system_mta(file):
        return _load_system(file)
    else:
        raise ValueError(f"File does not contain a valid System object: {file}")


def _is_system_mta(path: Union[str, Path, io.BytesIO]) -> bool:
    zipf = zipfile.ZipFile(path, "r")
    all_zip_files = zipf.namelist()
    required_files = ["positions.npy", "cell.npy", "types.npy", "pbc.npy"]
    for file in required_files:
        if file not in all_zip_files:
            return False
    return True


def _load_system(path: Union[str, Path, io.BytesIO]) -> System:
    # we filter a warning related to the fact that numpy arrays from buffers
    # are not writable, while torch would like arrays to be writable when
    # converting them to tensors; this is ok because we then clone the tensor
    warnings.filterwarnings(
        "ignore", category=UserWarning, message="The given NumPy array is not writable"
    )

    with zipfile.ZipFile(path, "r") as zipf:
        positions = torch.from_numpy(np.load(zipf.open("positions.npy")))
        cell = torch.from_numpy(np.load(zipf.open("cell.npy")))
        types = torch.from_numpy(np.load(zipf.open("types.npy")))
        pbc = torch.from_numpy(np.load(zipf.open("pbc.npy")))

        neighbor_list_options_list = []
        neighbor_lists = []
        for path in zipf.namelist():
            if path.startswith("pairs/") and path.endswith("/options.json"):
                nl_options = NeighborListOptions(0.0, False, False)
                nl_options._get_method("__setstate__")(zipf.read(path))
                neighbor_list_options_list.append(nl_options)

                data_path = path[:-12] + "data.mts"
                numpy_buffer = np.frombuffer(zipf.read(data_path), dtype=np.uint8)
                tensor_buffer = torch.from_numpy(numpy_buffer)
                neighbor_lists.append(metatensor.torch.load_block_buffer(tensor_buffer))

        extra_data_dict = {}
        for path in zipf.namelist():
            if path.startswith("data/"):
                name = os.path.basename(path).replace(".mts", "")
                numpy_buffer = np.frombuffer(zipf.read(path), dtype=np.uint8)
                tensor_buffer = torch.from_numpy(numpy_buffer)
                extra_data_dict[name] = metatensor.torch.load_buffer(tensor_buffer)

    system = System(
        positions=positions,
        cell=cell,
        types=types,
        pbc=pbc,
    )

    for options, neighbors in zip(neighbor_list_options_list, neighbor_lists):
        system.add_neighbor_list(options, neighbors)

    for key, value in extra_data_dict.items():
        system.add_data(key, value)

    return system


def _is_system(data: torch.ScriptObject) -> bool:
    if not isinstance(data, torch.ScriptObject):
        return False
    try:
        data.positions
        data.cell
        data.types
        data.pbc
        return True
    except AttributeError:
        return False


def _save_system(path: Union[str, Path], system: System) -> None:
    with zipfile.ZipFile(path, "w") as zipf:
        for nl_idx, nl_options in enumerate(system.known_neighbor_lists()):
            zipf.writestr(f"pairs/{nl_idx}/options.json", nl_options.__getstate__()[0])
            nl = system.get_neighbor_list(nl_options)
            tensor_buffer = metatensor.torch.save_buffer(nl)
            zipf.writestr(f"pairs/{nl_idx}/data.mts", tensor_buffer.numpy().tobytes())

        for key in system.known_data():
            data = system.get_data(key)
            tensor_buffer = metatensor.torch.save_buffer(data)
            zipf.writestr(f"data/{key}.mts", tensor_buffer.numpy().tobytes())

        for tensor_name in ["positions", "cell", "types", "pbc"]:
            tensor = getattr(system, tensor_name)
            numpy_array = tensor.numpy()
            with zipf.open(f"{tensor_name}.npy", "w") as fd:
                np.save(fd, numpy_array)



from typing import Dict, List, Optional

import torch
from metatensor.torch import Labels, TensorBlock, TensorMap, dtype_name

from . import ModelOutput, System


def _check_outputs(
    systems: List[System],
    requested: Dict[str, ModelOutput],
    selected_atoms: Optional[Labels],
    outputs: Dict[str, TensorMap],
    expected_dtype: torch.dtype,
):
    """
    Check that the outputs of a model conform to the expected structure for atomistic
    models.

    This function checks conformance with the reference documentation in
    https://docs.metatensor.org/metatomic/latest/outputs/index.html
    """

    for name, output in outputs.items():
        if requested.get(name) is None:
            raise ValueError(
                f"the model produced an output named '{name}', which was not requested"
            )

        if len(output) != 0:
            output_dtype = output.block_by_id(0).values.dtype
            if output_dtype != expected_dtype:
                raise ValueError(
                    f"wrong dtype for the {name} output: "
                    f"the model promised {dtype_name(expected_dtype)}, "
                    f"we got {dtype_name(output_dtype)}"
                )

    for name, request in requested.items():
        value = outputs.get(name)
        if value is None:
            raise ValueError(
                f"the model did not produce the '{name}' output, which was requested"
            )

        if name in ["energy", "energy_ensemble", "energy_uncertainty"]:
            _check_energy_like(name, value, systems, request, selected_atoms)
        elif name == "features":
            _check_features(value, systems, request, selected_atoms)
        elif name == "non_conservative_forces":
            _check_non_conservative_forces(value, systems, request, selected_atoms)
        elif name == "non_conservative_stress":
            _check_non_conservative_stress(value, systems, request)
        else:
            # this is a non-standard output, there is nothing to check
            continue


def _check_energy_like(
    name: str,
    value: TensorMap,
    systems: List[System],
    request: ModelOutput,
    selected_atoms: Optional[Labels],
):
    """
    Check the output metadata of energy-related outputs
    """
    assert name in ["energy", "energy_ensemble", "energy_uncertainty"]

    # Ensure the output contains a single block with the expected key
    _validate_single_block(name, value)

    # Check samples values from systems & selected_atoms
    _validate_atomic_samples(name, value, systems, request, selected_atoms)

    energy_block = value.block_by_id(0)
    device = value.device

    # Ensure that the block has no components
    _validate_no_components(name, energy_block)

    # The only difference between energy & energy_ensemble is in the properties
    if name == "energy" or name == "energy_uncertainty":
        expected_properties = Labels("energy", torch.tensor([[0]], device=device))
        message = "`Labels('energy', [[0]])`"
    else:
        assert name == "energy_ensemble"
        n_ensemble_members = energy_block.values.shape[-1]
        expected_properties = Labels(
            "energy", torch.arange(n_ensemble_members, device=device).reshape(-1, 1)
        )
        message = "`Labels('energy', [[0], ..., [n]])`"

    if energy_block.properties != expected_properties:
        raise ValueError(f"invalid properties for '{name}' output: expected {message}")

    for parameter, gradient in energy_block.gradients():
        if parameter not in ["strain", "positions"]:
            raise ValueError(f"invalid gradient for '{name}' output: {parameter}")

        xyz = torch.tensor([[0], [1], [2]], device=device)
        # strain gradient checks
        if parameter == "strain":
            if gradient.samples.names != ["sample"]:
                raise ValueError(
                    f"invalid samples for '{name}' output 'strain' gradients: "
                    f"expected the names to be ['sample'], got {gradient.samples.names}"
                )

            if len(gradient.components) != 2:
                raise ValueError(
                    f"invalid components for '{name}' output 'strain' gradients: "
                    "expected two components"
                )

            if gradient.components[0] != Labels("xyz_1", xyz):
                raise ValueError(
                    f"invalid components for '{name}' output 'strain' gradients: "
                    "expected Labels('xyz_1', [[0], [1], [2]]) for the first component"
                )

            if gradient.components[1] != Labels("xyz_2", xyz):
                raise ValueError(
                    f"invalid components for '{name}' output 'strain' gradients: "
                    "expected Labels('xyz_2', [[0], [1], [2]]) for the second component"
                )

        # positions gradient checks
        if parameter == "positions":
            if gradient.samples.names != ["sample", "system", "atom"]:
                raise ValueError(
                    f"invalid samples for '{name}' output 'positions' gradients: "
                    "expected the names to be ['sample', 'system', 'atom'], "
                    f"got {gradient.samples.names}"
                )

            if len(gradient.components) != 1:
                raise ValueError(
                    f"invalid components for '{name}' output 'positions' gradients: "
                    "expected one component"
                )

            if gradient.components[0] != Labels("xyz", xyz):
                raise ValueError(
                    f"invalid components for '{name}' output 'positions' gradients: "
                    "expected Labels('xyz', [[0], [1], [2]]) for the first component"
                )


def _check_features(
    value: TensorMap,
    systems: List[System],
    request: ModelOutput,
    selected_atoms: Optional[Labels],
):
    """
    Check "features" output metadata. It is standardized with Plumed
    https://www.plumed.org/doc-master/user-doc/html/_m_e_t_a_t_e_n_s_o_r.html
    """
    # Ensure the output contains a single block with the expected key
    _validate_single_block("features", value)

    # Check samples values from systems & selected_atoms
    _validate_atomic_samples("features", value, systems, request, selected_atoms)

    features_block = value.block_by_id(0)

    # Check that the block has no components
    _validate_no_components("features", features_block)

    # Should not have any explicit gradients
    # all gradient calculations are done using autograd
    if len(features_block.gradients_list()) > 0:
        raise ValueError(
            "invalid gradients for 'features' output: it should not have any explicit",
            "gradients. all gradient calculations should be done using autograd",
        )


def _check_non_conservative_forces(
    value: TensorMap,
    systems: List[System],
    request: ModelOutput,
    selected_atoms: Optional[Labels],
):
    """
    Check output metadata for non-conservative forces.
    """
    # Ensure the output contains a single block with the expected key
    _validate_single_block("non_conservative_forces", value)

    # Check samples values from systems & selected_atoms
    _validate_atomic_samples(
        "non_conservative_forces", value, systems, request, selected_atoms
    )

    forces_block = value.block_by_id(0)

    # Check that the block has correct "Cartesian-form" components
    if len(forces_block.components) != 1:
        raise ValueError(
            "invalid components for 'non_conservative_forces' output: "
            "expected one component"
        )
    expected_component = Labels(
        "xyz", torch.tensor([[0], [1], [2]], device=value.device)
    )
    if forces_block.components[0] != expected_component:
        raise ValueError(
            f"invalid components for 'non_conservative_forces' output: "
            f"expected {expected_component}, got {forces_block.components[0]}"
        )

    # Should not have any gradients
    if len(forces_block.gradients_list()) > 0:
        raise ValueError(
            "invalid gradients for 'non_conservative_forces' output: "
            f"expected no gradients, found {forces_block.gradients_list()}"
        )


def _check_non_conservative_stress(
    value: TensorMap,
    systems: List[System],
    request: ModelOutput,
):
    """
    Check output metadata for the non-conservative stress.
    """
    # Ensure the output contains a single block with the expected key
    _validate_single_block("non_conservative_stress", value)

    # Check samples values from systems
    _validate_atomic_samples(
        "non_conservative_stress", value, systems, request, selected_atoms=None
    )

    stress_block = value.block_by_id(0)

    # Check that the block has correct "Cartesian-form" components
    if len(stress_block.components) != 2:
        raise ValueError(
            "invalid components for 'non_conservative_stress' output: "
            f"expected two components, got {len(stress_block.components)}"
        )
    expected_components = [
        Labels("xyz_1", torch.tensor([[0], [1], [2]], device=value.device)),
        Labels("xyz_2", torch.tensor([[0], [1], [2]], device=value.device)),
    ]
    for expected, actual in zip(expected_components, stress_block.components):
        if expected != actual:
            raise ValueError(
                f"invalid components for 'non_conservative_stress' output: "
                f"expected {expected}, got {actual}"
            )

    # Should not have any gradients
    if len(stress_block.gradients_list()) > 0:
        raise ValueError(
            "invalid gradients for 'non_conservative_stress' output: "
            f"expected no gradients, found {stress_block.gradients_list()}"
        )


def _validate_atomic_samples(
    name: str,
    value: TensorMap,
    systems: List[System],
    request: ModelOutput,
    selected_atoms: Optional[Labels],
):
    """
    Validates the sample labels in the output against the expected structure
    """
    device = value.device
    block = value.block_by_id(0)

    # Check if the samples names are as expected based on whether the output is
    # per-atom or global
    if request.per_atom:
        expected_samples_names = ["system", "atom"]
    else:
        expected_samples_names = ["system"]

    if block.samples.names != expected_samples_names:
        raise ValueError(
            f"invalid sample names for '{name}' output: "
            f"expected {expected_samples_names}, got {block.samples.names}"
        )

    # Check if the samples match the systems and selected_atoms
    if request.per_atom:
        expected_values: List[List[int]] = []
        for s, system in enumerate(systems):
            for a in range(len(system)):
                expected_values.append([s, a])
        expected_samples = Labels(
            ["system", "atom"], torch.tensor(expected_values, device=device)
        )
        if selected_atoms is not None:
            expected_samples = expected_samples.intersection(selected_atoms)
    else:
        expected_samples = Labels(
            "system", torch.arange(len(systems), device=device).reshape(-1, 1)
        )
        if selected_atoms is not None:
            selected_systems = Labels(
                "system", torch.unique(selected_atoms.column("system")).reshape(-1, 1)
            )
            expected_samples = expected_samples.intersection(selected_systems)

    if len(expected_samples.union(block.samples)) != len(expected_samples):
        raise ValueError(
            f"invalid samples entries for '{name}' output, they do not match the "
            f"`systems` and `selected_atoms`. Expected samples:\n{expected_samples}"
        )


def _validate_single_block(name: str, value: TensorMap):
    """
    Ensure the TensorMap has a single block with the expected key
    """
    if value.keys != Labels("_", torch.tensor([[0]])):
        raise ValueError(
            f"invalid keys for '{name}' output: expected `Labels('_', [[0]])`"
        )


def _validate_no_components(name: str, block: TensorBlock):
    """
    Ensure the block has no components"
    """
    if len(block.components) != 0:
        raise ValueError(
            f"invalid components for {name} output: components should be empty"
        )



import warnings
from typing import List, Optional, Union

import numpy as np
import torch

from . import System


try:
    import ase

    HAS_ASE = True
except ImportError:
    HAS_ASE = False


class IntoSystem:
    """
    A type that can be converted into a :py:class:`metatomic.torch.System`.

    This is an abstract class that is used to indicate a class whose objects can be
    converted into a :py:class:`System`. For the moment, the only supported type is
    :py:class:`ase.Atoms`.
    """


def systems_to_torch(
    systems: Union[IntoSystem, List[IntoSystem]],
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
    positions_requires_grad: bool = False,
    cell_requires_grad: bool = False,
) -> Union[System, List[System]]:
    """Converts a system or a list of systems into a
    ``metatomic.torch.System`` or a list of such objects.

    :param: systems: The system or list of systems to convert.
    :param: dtype: The dtype of the output tensors. If ``None``, the default
        dtype is used.
    :param: device: The device of the output tensors. If ``None``, the default
        device is used.
    :param: positions_requires_grad: Whether the positions tensors of
        the outputs should require gradients.
    :param: cell_requires_grad: Whether the cell tensors of the outputs
        should require gradients.

    :return: The converted system or list of systems.
    """

    if isinstance(systems, list):
        return [
            _system_to_torch(
                system, dtype, device, positions_requires_grad, cell_requires_grad
            )
            for system in systems
        ]
    else:
        return _system_to_torch(
            systems, dtype, device, positions_requires_grad, cell_requires_grad
        )


def _system_to_torch(
    system: IntoSystem,
    dtype: Optional[torch.dtype],
    device: Optional[torch.device],
    positions_requires_grad: bool,
    cell_requires_grad: bool,
) -> System:
    """Converts a system into a ``metatomic.torch.System``.

    :param: system: The system to convert.
    :param: dtype: The dtype of the output tensors. If ``None``, the default
        dtype is used.
    :param: device: The device of the output tensors. If ``None``, the default
        device is used.
    :param: positions_requires_grad: Whether the positions tensors of
        the outputs should require gradients.
    :param: cell_requires_grad: Whether the cell tensors of the outputs
        should require gradients.

    :return: The converted system.
    """
    if not HAS_ASE:
        raise RuntimeError("The `ase` package is required to convert systems to torch.")

    if not isinstance(system, ase.Atoms):
        raise ValueError(
            "Only `ase.Atoms` objects can be converted to `System`s "
            f"for now; got {type(system)}."
        )

    if dtype is None:
        # this is necessary because creating torch tensors from numpy arrays
        # takes the dtype from the numpy array, which is not always the default
        # dtype
        dtype = torch.get_default_dtype()

    positions = torch.tensor(
        system.positions,
        requires_grad=positions_requires_grad,
        dtype=dtype,
        device=device,
    )

    cell_vectors_are_not_zero = np.any(system.cell != 0, axis=1)
    if not np.all(cell_vectors_are_not_zero == system.pbc):
        warnings.warn(
            "A conversion to `System` was requested for an `ase.Atoms` object "
            "with one or more non-zero cell vectors but where the corresponding "
            "boundary conditions are set to `False`. "
            "The corresponding cell vectors will be set to zero.",
            stacklevel=3,
        )

    cell = torch.zeros((3, 3), dtype=dtype, device=device)

    pbc = torch.tensor(system.pbc, dtype=torch.bool, device=device)

    cell[pbc] = torch.tensor(system.cell[system.pbc], dtype=dtype, device=device)

    types = torch.tensor(system.numbers, device=device, dtype=torch.int32)

    return System(positions=positions, cell=cell, types=types, pbc=pbc)
